//---------------------------------------------------------------------
// SPACE TAPPER: effects.c
// by Omar Cornut
//---------------------------------------------------------------------
// Effects
//---------------------------------------------------------------------

#include "tapper.h"

//---------------------------------------------------------------------
// DATA
//---------------------------------------------------------------------

static int		value;
static int		effect_type;
static int		effect_time;
static int		effect_args[4];

static int		effects_close = YES;
static int		(*effects_close_callback)(void);

//---------------------------------------------------------------------
// FUNCTIONS
//---------------------------------------------------------------------

// If effect_close is disabled, then the closing stuff will not be performed 
// at the end of effect. It is useful if you want to chain different similar 
// effects, without disabling an hardware feature in-between
// In addition, a callback can be provided as well. If the callback return FALSE
// the closure will not be performed as well.
// Note: if provided, the callback is called even if close == FALSE
void		effects_setup_closure(int close, int (*callback)(void))
{
	effects_close = close;
	effects_close_callback = callback;
}

//---------------------------------------------------------------------

void			effects_start(va_list args)
{
	int			target;
	
	value = 0;
	effect_type = va_arg(args, int);
	effect_time = va_arg(args, int) - 1;
	
	switch (effect_type)
	{
		case EFFECT_P0_FADE_IN:
		{
			target = va_arg(args, int);
			REG_BLDCNT = target | REG_BLDCNT_EFFECT_BRIGHTNESS_DEC;
			REG_BLDY = REG_BLDY_EVY_Set(31);
			break;
		}
		case EFFECT_P0_FADE_OUT:
		{
			target = va_arg(args, int);
			REG_BLDCNT = target | REG_BLDCNT_EFFECT_BRIGHTNESS_DEC;
			REG_BLDY = REG_BLDY_EVY_Set(0);
			break;
		}
		case EFFECT_P0_MOSAIC_IN:
		{
			game.bg0.cnt |= REG_BGxCNT_MOSAIC;
			background_sync_cnt(&game.bg0);
			REG_MOSAIC = REG_MOSAIC_BG_SIZE_Set(15);
			break;
		}
		case EFFECT_P0_MOSAIC_OUT:
		{
			game.bg0.cnt |= REG_BGxCNT_MOSAIC;
			background_sync_cnt(&game.bg0);
			REG_MOSAIC = REG_MOSAIC_BG_SIZE_Set(0);
			break;
		}
		case EFFECT_P0_WINDOW_IN:
		{
			windows.win0.coord.x1 = effect_args[0] = va_arg(args, int);
			windows.win0.coord.y1 = effect_args[1] = va_arg(args, int);
			windows.win0.coord.x2 = effect_args[2] = va_arg(args, int);
			windows.win0.coord.y2 = effect_args[3] = va_arg(args, int);
	    	windows.win0.in.reg   = REG_WININ_DISP_ALL;
    		windows.wins_out.reg  = REG_WINOUT_WIN01_DISP_NONE;
		    windows_sync();
		    REG_DISPCNT |= REG_DISPCNT_WIN0_ENABLE;
			break;
		}
		case EFFECT_P0_WINDOW_OUT:
		{
			effect_args[0] = va_arg(args, int);
			effect_args[1] = va_arg(args, int);
			effect_args[2] = va_arg(args, int);
			effect_args[3] = va_arg(args, int);
			windows.win0.coord.x1 = 0;
			windows.win0.coord.y1 = 0;
			windows.win0.coord.x2 = RES_WIDTH - 1;
			windows.win0.coord.y2 = RES_HEIGHT - 1;
	    	windows.win0.in.reg   = REG_WININ_DISP_ALL;
    		windows.wins_out.reg  = REG_WINOUT_WIN01_DISP_NONE;
		    windows_sync();
		    REG_DISPCNT |= REG_DISPCNT_WIN0_ENABLE;
			break;
		}
	}
}

void			effects_update(void)
{
	switch (effect_type)
	{
		case EFFECT_P0_FADE_IN:
		{
			REG_BLDY = REG_BLDY_EVY_Set(INTERPOLATE(31, 0, effect_time, value));
			if (value++ == effect_time)
				section_return();
			break;
		}
		case EFFECT_P0_FADE_OUT:
		{
			REG_BLDY = REG_BLDY_EVY_Set(INTERPOLATE(0, 31, effect_time, value));
			if (value++ == effect_time)
				section_return();
			break;
		}
		case EFFECT_P0_MOSAIC_IN:
		{
			REG_MOSAIC = REG_MOSAIC_BG_SIZE_Set(INTERPOLATE(15, 0, effect_time, value));
			if (value++ == effect_time)
				section_return();
			break;
		}
		case EFFECT_P0_MOSAIC_OUT:
		{
			REG_MOSAIC = REG_MOSAIC_BG_SIZE_Set(INTERPOLATE(0, 15, effect_time, value));
			if (value++ == effect_time)
				section_return();
			break;
		}
		case EFFECT_P0_WINDOW_IN:
		{
			windows.win0.coord.x1 = INTERPOLATE(effect_args[0], 0,              effect_time, value);
        	windows.win0.coord.y1 = INTERPOLATE(effect_args[1], 0,              effect_time, value);
            windows.win0.coord.x2 = INTERPOLATE(effect_args[2], RES_WIDTH  - 1, effect_time, value);
            windows.win0.coord.y2 = INTERPOLATE(effect_args[3], RES_HEIGHT - 1, effect_time, value);
			windows_sync();                
			if (value++ == effect_time)
				section_return();
			break;
		}
		case EFFECT_P0_WINDOW_OUT:
		{
			windows.win0.coord.x1 = INTERPOLATE(0,              effect_args[0], effect_time, value);
        	windows.win0.coord.y1 = INTERPOLATE(0,              effect_args[1], effect_time, value);
            windows.win0.coord.x2 = INTERPOLATE(RES_WIDTH  - 1, effect_args[2], effect_time, value);
            windows.win0.coord.y2 = INTERPOLATE(RES_HEIGHT - 1, effect_args[3], effect_time, value);
			windows_sync();                
			if (value++ == effect_time)
				section_return();
			break;
		}
	}
}

void			effects_stop(void)
{
	if (effects_close_callback != NULL && effects_close_callback() == FALSE)
		return;		
	if (effects_close == FALSE)
		return;
	switch (effect_type)
	{
		case EFFECT_P0_FADE_IN:
		case EFFECT_P0_FADE_OUT:
		{
			REG_BLDCNT = 0x0000;
			break;
		}
		case EFFECT_P0_MOSAIC_IN:
		case EFFECT_P0_MOSAIC_OUT:
		{
			game.bg0.cnt &= ~REG_BGxCNT_MOSAIC;
			background_sync_cnt(&game.bg0);
			break;
		}
		case EFFECT_P0_WINDOW_IN:
		case EFFECT_P0_WINDOW_OUT:
		{
			REG_DISPCNT &= ~REG_DISPCNT_WIN0_ENABLE;
			break;
		}
	}
}

//---------------------------------------------------------------------
