//---------------------------------------------------------------------
// GBALIB: anim.c
// by Omar Cornut
//---------------------------------------------------------------------
// Animation System
//---------------------------------------------------------------------
// A t_animator is an object playing animations based on an animation set.
// Typically, each moving sprite/character within a game could have an
// associated t_animator.
// An animation set is static (defined in ROM) and contains a table of
// animations, which are themselves	a list of pictures with associated,
// delay position and bounding box information.
// Note: the animation system is UNSAFE and avoid controlling data for
// performance purpose. Please insure the correctness of your data.
//---------------------------------------------------------------------
// Needs:
//	- boundingbox
//	- sprites
//	- DMA
//---------------------------------------------------------------------

#include "gbalib.h"

//---------------------------------------------------------------------

// FUNCTION: animator_init()
// Initialize an animator
void		animator_init(t_animator *a, ts_anims_set *anims_set, t_sprite *sprite, int mode, void *arg1, void *arg2)
{
	// Variable (State)
	a->state.loop_count = 0;
	a->state.delay_current = 1;
	a->state.anim_current = 0;
	a->state.frame_current = NULL;
	// Fixed
	a->mode = mode;
	a->anims_set = anims_set;
	a->sprite = sprite;
	switch (mode)
	{
		case ANIMATOR_MODE_OAM_CHANGE_IDX:
		{
			a->mode_oam_change_idx.idx_start = (u16)(u32)arg1;
			a->mode_oam_change_idx.idx_increment = anims_set->picture_size / TILE_SIZE_16; /* FIXME: 256 colors sprites, how tiles should be addressed ? */
			break;
		}
		case ANIMATOR_MODE_VRAM_COPY_DATA:
		{
			a->mode_vram_copy_data.src_rom_tab = (u16 *)arg1;
			a->mode_vram_copy_data.dst_video = (u16 *)arg2;
			break;
		}
	}
}

// [PRIVATE] FUNCTION: animator_video_update()
// Synchronize current frame with display, either by changing a sprite idx, or copying DMA data, etc..
// FIXME: avoid synchronizing if display has already the same data
static void	animator_video_update(t_animator *a)
{
	ts_anim_frame *frame = a->state.frame_current;
	switch (a->mode)
	{
		case ANIMATOR_MODE_OAM_CHANGE_IDX:
		{
			a->sprite->oam.charno = a->mode_oam_change_idx.idx_start + (frame->i_picture * a->mode_oam_change_idx.idx_increment);
			break;
		}
		case ANIMATOR_MODE_VRAM_COPY_DATA:
		{
			DMA_Copy(3, DMA_32_NOW,
					(void *)((u8 *)a->mode_vram_copy_data.src_rom_tab + (frame->i_picture * a->anims_set->picture_size)),
					a->mode_vram_copy_data.dst_video, 
					a->anims_set->picture_size / sizeof (u32));
			break;
		}
	}
	// Update offset if there's an associated sprite
	if (a->sprite)
	{
		u16 offsets = *(u16 *)((u8 *)a->anims_set + a->anims_set->positions_offset + frame->i_position * sizeof(ts_anim_position));
		if (a->sprite->oam.hflip)
			offsets = (offsets & 0xFF00) | (-offsets & 0x00FF);
		a->sprite->offsets = offsets;
	}
}

// FUNCTION: animator_start()
// Start given animation at given frame
void		animator_start(t_animator *a, int anim_index, int frame_index)
{
	// Current Anim
	ts_anim *anim = (ts_anim *)((u8 *)a->anims_set + a->anims_set->anims_offsets[anim_index]);

	// Reset Loop Counter	
	a->state.loop_count = 0;
	
	// Set given Anim
	a->state.anim_current = anim_index;
	
	// Set given Frame
	a->state.frame_current = &anim->frames + frame_index;
	
	// Setup new Delay
	a->state.delay_current = a->state.frame_current->delay;

	// Setup current frame for display
	animator_video_update(a);
}

// FUNCTION: animator_set()
// Set given animation at given frame
void		animator_set(t_animator *a, int anim_index, int frame_index)
{
	// Current Anim
	ts_anim *anim = (ts_anim *)((u8 *)a->anims_set + a->anims_set->anims_offsets[anim_index]);

	// Reset Loop Counter	
	a->state.loop_count = 0;
	
	// Set given Anim
	a->state.anim_current = anim_index;
	
	// Set given Frame
	a->state.frame_current = &anim->frames + frame_index;
	
	// Setup new Delay
	a->state.delay_current = a->state.frame_current->delay;
}

// FUNCTION: animator_set_frame()
// Set current animation to given frame
void		animator_set_frame(t_animator *a, int frame_index)
{
	// Current Anim
	ts_anim *anim = (ts_anim *)((u8 *)a->anims_set + a->anims_set->anims_offsets[a->state.anim_current]);

	// Switch to Given Frame
	a->state.frame_current = &anim->frames + frame_index;

	// Setup new Delay
	a->state.delay_current = a->state.frame_current->delay;

	// Setup new frame for display
	animator_video_update(a);
}

// FUNCTION: animator_update()
// Update the animator (play current animation, etc..)
// Must be called each time unit (usually 60 times per seconds)
void		animator_update(t_animator *a)
{
	if (--a->state.delay_current > 0)
		return;
		
	// Current Anim
	ts_anim *anim = (ts_anim *)((u8 *)a->anims_set + a->anims_set->anims_offsets[a->state.anim_current]);
	
	// Switch to Next Frame
	a->state.frame_current++;

	// Check delay
	// If delay == 0, this is the Last Frame of animation
	if (a->state.frame_current->delay == 0)
	{
		// Switch to Next Animation
		a->state.anim_current = anim->next_anim;
		
		// Set to Looping Frame of Animation
		ts_anim *old_anim = anim;
		anim = (ts_anim *)((u8 *)a->anims_set + a->anims_set->anims_offsets[a->state.anim_current]);
		a->state.frame_current = &anim->frames + old_anim->next_frame;

		// Increment Loop Counter
		a->state.loop_count++;
	}
	
	// Setup new Delay
	a->state.delay_current = a->state.frame_current->delay;
	
	// Setup new frame for display
	animator_video_update(a);
}

//---------------------------------------------------------------------

// FUNCTION: animator_get_current_boundingbox()
// Return address to the boundingbox data for the current animation frame
t_boundingbox *animator_get_current_boundingbox(t_animator *a)
{
	t_boundingbox *bb;
	bb = (t_boundingbox *)((u8 *)a->anims_set + a->anims_set->boundingboxes_offset) + a->state.frame_current->i_boundingbox * ANIMATOR_BOUNDINGBOX_FLIP_MAX;
	return (bb);
}

//---------------------------------------------------------------------

// FUNCTION: animator_state_backup()
// Backup an animator state
void	animator_state_backup(t_animator *src, t_animator_state *dst)
{
	dst->loop_count = src->state.loop_count;
	dst->delay_current = src->state.delay_current;
	dst->anim_current = src->state.anim_current;
	dst->frame_current = src->state.frame_current;
}

// FUNCTION: animator_state_restore()
// Restore an animator state
void	animator_state_restore(t_animator *dst, t_animator_state *src)
{
	dst->state.loop_count = src->loop_count;
	dst->state.delay_current = src->delay_current;
	dst->state.anim_current = src->anim_current;
	dst->state.frame_current = src->frame_current;

	// Setup current frame for display
	animator_video_update(dst);
}

//---------------------------------------------------------------------
