//---------------------------------------------------------------------
// GBALIB: gba.h
// by Omar Cornut
//---------------------------------------------------------------------
// Basic Types
// Basic Macros
// GBA Registers
// GBA Definitions
//---------------------------------------------------------------------

// Define GBALIB_SAFE to enable masking in _Set() macros
// #define GBALIB_SAFE

//---------------------------------------------------------------------
// BASIC TYPES
//---------------------------------------------------------------------

typedef	unsigned char		u8;
typedef	unsigned short		u16;
typedef	unsigned long		u32;

typedef	signed char			s8;
typedef	signed short		s16;
typedef	signed long			s32;

//---------------------------------------------------------------------
// BASIC MACROS
//---------------------------------------------------------------------

// NULL
#ifndef	NULL
#define	NULL				((void *)0)
#endif

// TRUE/FALSE, YES/NO, ON/OFF
#define	TRUE				(1)
#define	FALSE				(0)
#define	YES					(1)
#define	NO					(0)
#define	ON					(1)
#define	OFF					(0)

// Sub-macro to	use	for	macros replacing a code	statement
#define	MACRO_START			do {
#define	MACRO_END			} while	(0)

// Linear Inperpolation
#define	INTERPOLATE(BEGIN,END,STEPS,CURRENT)				\
		(BEGIN + ((((END)-(BEGIN)) * (CURRENT))	/ STEPS))

// Setup MASK macro
#ifdef GBALIB_SAFE
#define GBALIB_MASK(_V)		& (_V)
#else
#define GBALIB_MASK(_V)
#endif

//---------------------------------------------------------------------
// GBA MACROS
//---------------------------------------------------------------------

// Resolutions
#define	RES_WIDTH			(240)				/* Screen Width									*/
#define	RES_HEIGHT			(160)				/* Screen Height								*/
#define	RES_WIDTH_TILES		(30)				/* Screen Width	in Tiles						*/
#define	RES_HEIGHT_TILES	(20)				/* Screen Height in	Tiles						*/
#define RES_WIDTH_MAP		(32)				/* Mapping Width in Tiles						*/
#define RES_HEIGHT_MAP		(32)				/* Mapping Height in Tiles						*/
#define	RES_WIDTH_VDP		(308)				/* VDP Width (outside of REG_WIDTH is H-Blank	*/
#define	RES_HEIGHT_VDP		(228)				/* VDP Height (outside of REG_HEIGHT is	V-Blank	*/

// Convert RGB components (from	0 to 31) to	a 15-bit BGR value
#define	RGB16(r,g,b)		((r) | ((g)	<< 5) |	((b) <<	10))

// Tile	Sizes
#define	TILE_SIZE_16		((8*8)/2)			/* 32 bytes	*/
#define	TILE_SIZE_256		(8*8)				/* 64 bytes	*/

//---------------------------------------------------------------------
// REGIONS ON VIRTUAL MEMORY MAP
//---------------------------------------------------------------------

#define	MEM_SYSROM			((u16*)0x0000000)	/* System ROM									*/
#define	MEM_EXRAM			((u16*)0x2000000)	/* External	WRAM (256 kb)						*/
#define	MEM_RAM				((u16*)0x3000000)	/* Internal	RAM	(32	kb)							*/
#define	MEM_IO				((u16*)0x4000000)	/* Registers									*/
#define	MEM_PAL				((u16*)0x5000000)	/* Palette Memory								*/
#define	MEM_PAL_BG			((u16*)0x5000000)	/* Palette Memory (Background)					*/
#define	MEM_PAL_OBJ			((u16*)0x5000200)	/* Palette Memory (Sprites)						*/
#define	MEM_VRAM			((u16*)0x6000000)	/* Video RAM (96 kb)							*/
#define	MEM_VRAM_BG			((u16*)0x6000000)	/* Video RAM, Background Tiles Area				*/
#define	MEM_VRAM_OBJ		((u16*)0x6010000)	/* Video RAM, Sprites Tiles	Area				*/
#define	MEM_OAM				((u32*)0x7000000)	/* OAM (Sprite Memory)							*/
#define	MEM_ROM				((u16*)0x8000000)	/* ROM (GamePak)								*/
#define	MEM_SRAM			((u16*)0xE000000)	/* SaveRAM (located	on GamePak)					*/

#define	MEM_PALx_BG(n)			(MEM_PAL_BG+((n)*16))					/* Pointer to given	Background Palette	*/
#define	MEM_PALx_OBJ(n)			(MEM_PAL_OBJ+((n)*16))					/* Pointer to given	Sprites	Palette		*/
#define	MEM_VRAM_BG_TILEx(n)	(MEM_VRAM_BG+((n)*(TILE_SIZE_16/2)))	/* Pointer to given	Background Tile		*/
#define	MEM_VRAM_OBJ_TILEx(n)	(MEM_VRAM_OBJ+((n)*(TILE_SIZE_16/2)))	/* Pointer to given	Sprites	Tile		*/
	
//---------------------------------------------------------------------
// HARDWARE	REGISTERS									ACCESS
//---------------------------------------------------------------------

/* Display */
#define	REG_DISPCNT			*(volatile u32*)0x4000000	/* RW */
#define	REG_DISPCNT_L		*(volatile u16*)0x4000000	/* RW */
#define	REG_DISPCNT_H		*(volatile u16*)0x4000002	/* RW */
#define	REG_DISPSTAT		*(volatile u16*)0x4000004	/* RW */
#define	REG_VCOUNT			*(volatile u16*)0x4000006	/* R. */

/* Backgrounds */
#define	REG_BG0CNT			*(volatile u16*)0x4000008	/* RW */
#define	REG_BG1CNT			*(volatile u16*)0x400000A	/* RW */
#define	REG_BG2CNT			*(volatile u16*)0x400000C	/* RW */
#define	REG_BG3CNT			*(volatile u16*)0x400000E	/* RW */
#define	REG_BGxCNT(x)		*(volatile u16*)(0x4000008+2*x)
#define	REG_BG0HOFS			*(volatile u16*)0x4000010	/* .W */
#define	REG_BG0VOFS			*(volatile u16*)0x4000012	/* .W */
#define	REG_BG1HOFS			*(volatile u16*)0x4000014	/* .W */
#define	REG_BG1VOFS			*(volatile u16*)0x4000016	/* .W */
#define	REG_BG2HOFS			*(volatile u16*)0x4000018	/* .W */
#define	REG_BG2VOFS			*(volatile u16*)0x400001A	/* .W */
#define	REG_BG3HOFS			*(volatile u16*)0x400001C	/* .W */
#define	REG_BG3VOFS			*(volatile u16*)0x400001E	/* .W */
#define	REG_BGxHOFS(x)		*(volatile u16*)(0x4000010+4*x)
#define	REG_BGxVOFS(x)		*(volatile u16*)(0x4000012+4*x)
#define	REG_BG2PA			*(volatile u16*)0x4000020	/* .W */
#define	REG_BG2PB			*(volatile u16*)0x4000022	/* .W */
#define	REG_BG2PC			*(volatile u16*)0x4000024	/* .W */
#define	REG_BG2PD			*(volatile u16*)0x4000026	/* .W */
#define	REG_BG2X			*(volatile u32*)0x4000028	/* .W */
#define	REG_BG2X_L			*(volatile u16*)0x4000028	/* .W */
#define	REG_BG2X_H			*(volatile u16*)0x400002A	/* .W */
#define	REG_BG2Y			*(volatile u32*)0x400002C	/* .W */
#define	REG_BG2Y_L			*(volatile u16*)0x400002C	/* .W */
#define	REG_BG2Y_H			*(volatile u16*)0x400002E	/* .W */
#define	REG_BG3PA			*(volatile u16*)0x4000030	/* .W */
#define	REG_BG3PB			*(volatile u16*)0x4000032	/* .W */
#define	REG_BG3PC			*(volatile u16*)0x4000034	/* .W */
#define	REG_BG3PD			*(volatile u16*)0x4000036	/* .W */
#define	REG_BG3X			*(volatile u32*)0x4000038	/* .W */
#define	REG_BG3X_L			*(volatile u16*)0x4000038	/* .W */
#define	REG_BG3X_H			*(volatile u16*)0x400003A	/* .W */
#define	REG_BG3Y			*(volatile u32*)0x400003C	/* .W */
#define	REG_BG3Y_L			*(volatile u16*)0x400003C	/* .W */
#define	REG_BG3Y_H			*(volatile u16*)0x400003E	/* .W */

/* Windows */
#define	REG_WIN0H			*(volatile u16*)0x4000040	/* .W */
#define	REG_WIN1H			*(volatile u16*)0x4000042	/* .W */
#define	REG_WIN0V			*(volatile u16*)0x4000044	/* .W */
#define	REG_WIN1V			*(volatile u16*)0x4000046	/* .W */
#define	REG_WININ			*(volatile u16*)0x4000048	/* RW */
#define	REG_WINOUT			*(volatile u16*)0x400004A	/* RW */

/* Mosaic */
#define	REG_MOSAIC			*(volatile u32*)0x400004C	/* .W */
#define	REG_MOSAIC_L		*(volatile u32*)0x400004C	/* .W */
#define	REG_MOSAIC_H		*(volatile u32*)0x400004E	/* .W */

/* Alpha Blending & Effects */
#define	REG_BLDCNT			*(volatile u16*)0x4000050	/* RW */
#define	REG_BLDALPHA		*(volatile u16*)0x4000052	/* RW */
#define	REG_BLDY			*(volatile u16*)0x4000054	/* .W */

/* Sound */
#define	REG_SND1CNT			*(volatile u32*)0x4000060	/* RW */
#define	REG_SND1CNT_L		*(volatile u16*)0x4000060	/* RW */
#define	REG_SND1CNT_H		*(volatile u16*)0x4000062	/* RW */
#define	REG_SND1CNT_X		*(volatile u16*)0x4000064	/* RW */
#define	REG_SND2CNT_L		*(volatile u16*)0x4000068	/* RW */
#define	REG_SND2CNT_H		*(volatile u16*)0x400006C	/* RW */
#define	REG_SND3CNT			*(volatile u32*)0x4000070	/* RW */
#define	REG_SND3CNT_L		*(volatile u16*)0x4000070	/* RW */
#define	REG_SND3CNT_H		*(volatile u16*)0x4000072	/* RW */
#define	REG_SND3CNT_X		*(volatile u16*)0x4000074	/* RW */
#define	REG_SND4CNT_L		*(volatile u16*)0x4000078	/* RW */
#define	REG_SND4CNT_H		*(volatile u16*)0x400007C	/* RW */
#define	REG_SNDCNT			*(volatile u32*)0x4000080	/* RW */
#define	REG_SNDCNT_L		*(volatile u16*)0x4000080	/* RW */
#define	REG_SNDCNT_H		*(volatile u16*)0x4000082	/* RW */
#define	REG_SNDCNT_X		*(volatile u16*)0x4000084	/* RW */
#define	REG_SNDBIAS			*(volatile u16*)0x4000088	/* RW */
#define	REG_WAVE_RAM0		*(volatile u32*)0x4000090	/* RW */
#define	REG_WAVE_RAM0_L		*(volatile u16*)0x4000090	/* RW */
#define	REG_WAVE_RAM0_H		*(volatile u16*)0x4000092	/* RW */
#define	REG_WAVE_RAM1		*(volatile u32*)0x4000094	/* RW */
#define	REG_WAVE_RAM1_L		*(volatile u16*)0x4000094	/* RW */
#define	REG_WAVE_RAM1_H		*(volatile u16*)0x4000096	/* RW */
#define	REG_WAVE_RAM2		*(volatile u32*)0x4000098	/* RW */
#define	REG_WAVE_RAM2_L		*(volatile u16*)0x4000098	/* RW */
#define	REG_WAVE_RAM2_H		*(volatile u16*)0x400009A	/* RW */
#define	REG_WAVE_RAM3		*(volatile u32*)0x400009C	/* RW */
#define	REG_WAVE_RAM3_L		*(volatile u16*)0x400009C	/* RW */
#define	REG_WAVE_RAM3_H		*(volatile u16*)0x400009E	/* RW */
#define	REG_FIFO_A			*(volatile u32*)0x40000A0	/* .W */
#define	REG_FIFO_A_L		*(volatile u16*)0x40000A0	/* .W */
#define	REG_FIFO_A_H		*(volatile u16*)0x40000A2	/* .W */
#define	REG_FIFO_B			*(volatile u32*)0x40000A4	/* .W */
#define	REG_FIFO_B_L		*(volatile u16*)0x40000A4	/* .W */
#define	REG_FIFO_B_H		*(volatile u16*)0x40000A6	/* .W */

/* DMA */
#define	REG_DMA0SAD			*(volatile u32*)0x40000B0	/* .W */
#define	REG_DMA0SAD_L		*(volatile u16*)0x40000B0	/* .W */
#define	REG_DMA0SAD_H		*(volatile u16*)0x40000B2	/* .W */
#define	REG_DMA0DAD			*(volatile u32*)0x40000B4	/* .W */
#define	REG_DMA0DAD_L		*(volatile u16*)0x40000B4	/* .W */
#define	REG_DMA0DAD_H		*(volatile u16*)0x40000B6	/* .W */
#define	REG_DMA0CNT			*(volatile u32*)0x40000B8	/* rW */
#define	REG_DMA0CNT_L		*(volatile u16*)0x40000B8	/* .W */
#define	REG_DMA0CNT_H		*(volatile u16*)0x40000BA	/* RW */
#define	REG_DMA1SAD			*(volatile u32*)0x40000BC	/* .W */
#define	REG_DMA1SAD_L		*(volatile u16*)0x40000BC	/* .W */
#define	REG_DMA1SAD_H		*(volatile u16*)0x40000BE	/* .W */
#define	REG_DMA1DAD			*(volatile u32*)0x40000C0	/* .W */
#define	REG_DMA1DAD_L		*(volatile u16*)0x40000C0	/* .W */
#define	REG_DMA1DAD_H		*(volatile u16*)0x40000C2	/* .W */
#define	REG_DMA1CNT			*(volatile u32*)0x40000C4	/* rW */
#define	REG_DMA1CNT_L		*(volatile u16*)0x40000C4	/* .W */
#define	REG_DMA1CNT_H		*(volatile u16*)0x40000C6	/* RW */
#define	REG_DMA2SAD			*(volatile u32*)0x40000C8	/* .W */
#define	REG_DMA2SAD_L		*(volatile u16*)0x40000C8	/* .W */
#define	REG_DMA2SAD_H		*(volatile u16*)0x40000CA	/* .W */
#define	REG_DMA2DAD			*(volatile u32*)0x40000CC	/* .W */
#define	REG_DMA2DAD_L		*(volatile u16*)0x40000CC	/* .W */
#define	REG_DMA2DAD_H		*(volatile u16*)0x40000CE	/* .W */
#define	REG_DMA2CNT			*(volatile u32*)0x40000D0	/* rW */
#define	REG_DMA2CNT_L		*(volatile u16*)0x40000D0	/* .W */
#define	REG_DMA2CNT_H		*(volatile u16*)0x40000D2	/* RW */
#define	REG_DMA3SAD			*(volatile u32*)0x40000D4	/* .W */
#define	REG_DMA3SAD_L		*(volatile u16*)0x40000D4	/* .W */
#define	REG_DMA3SAD_H		*(volatile u16*)0x40000D6	/* .W */
#define	REG_DMA3DAD			*(volatile u32*)0x40000D8	/* .W */
#define	REG_DMA3DAD_L		*(volatile u16*)0x40000D8	/* .W */
#define	REG_DMA3DAD_H		*(volatile u16*)0x40000DA	/* .W */
#define	REG_DMA3CNT			*(volatile u32*)0x40000DC	/* rW */
#define	REG_DMA3CNT_L		*(volatile u16*)0x40000DC	/* .W */
#define	REG_DMA3CNT_H		*(volatile u16*)0x40000DE	/* RW */

/* Timers */
#define	REG_TM0DATA			*(volatile u16*)0x4000100	/* RW */
#define	REG_TM0CNT_H		*(volatile u16*)0x4000102	/* RW */
#define	REG_TM1DATA			*(volatile u16*)0x4000104	/* RW */
#define	REG_TM1CNT_H		*(volatile u16*)0x4000106	/* RW */
#define	REG_TM2DATA			*(volatile u16*)0x4000108	/* RW */
#define	REG_TM2CNT_H		*(volatile u16*)0x400010A	/* RW */
#define	REG_TM3DATA			*(volatile u16*)0x400010C	/* RW */
#define	REG_TM3CNT_H		*(volatile u16*)0x400010E	/* RW */

/* SIO (Communication Port)	*/
#define	REG_RCNT			*(volatile u16*)0x4000134	/* RW */
#define	REG_SIOCNT			*(volatile u16*)0x4000128	/* rW */
#define	REG_SIODATA32_L		*(volatile u16*)0x4000120	/* RW */
#define	REG_SIODATA32_H		*(volatile u16*)0x4000122	/* RW */
#define	REG_SIODATA8		*(volatile u16*)0x400012A	/* RW */
#define	REG_SIOMULTI0		*(volatile u16*)0x4000120	/* RW */
#define	REG_SIOMULTI1		*(volatile u16*)0x4000122	/* RW */
#define	REG_SIOMULTI2		*(volatile u16*)0x4000124	/* RW */
#define	REG_SIOMULTI3		*(volatile u16*)0x4000126	/* RW */
#define	REG_SIOMULTI_SEND	*(volatile u16*)0x400012A	/* RW */

/* Communication / Inputs (JOY_BUS mode */
#define	REG_JOYCNT			*(volatile u16*)0x4000140	/* RW */
#define	REG_JOYRECV			*(volatile u32*)0x4000150	/* RW */
#define	REG_JOYRECV_L		*(volatile u16*)0x4000150	/* RW */
#define	REG_JOYRECV_H		*(volatile u16*)0x4000152	/* RW */
#define	REG_JOYTRANS		*(volatile u32*)0x4000154	/* RW */
#define	REG_JOYTRANS_L		*(volatile u16*)0x4000154	/* RW */
#define	REG_JOYTRANS_H		*(volatile u16*)0x4000156	/* RW */
#define	REG_JOYSTAT			*(volatile u32*)0x4000158	/* RW */
#define	REG_JOYSTAT_L		*(volatile u16*)0x4000158	/* RW */
#define	REG_JOYSTAT_H		*(volatile u16*)0x400015A	/* RW */

/* Inputs */
#define	REG_KEYINPUT		*(volatile u16*)0x4000130	/* RW */
#define	REG_KEYCNT			*(volatile u16*)0x4000132	/* RW */

/* Interrupts &	Miscellaneous */
#define	REG_INT_HANDLER		*(volatile u32*)0x3007FFC	/* RW */
#define	REG_IE				*(volatile u16*)0x4000200	/* RW */
#define	REG_IF				*(volatile u16*)0x4000202	/* RW */
#define	REG_WAITCNT			*(volatile u16*)0x4000204	/* RW */
#define	REG_IME				*(volatile u16*)0x4000208	/* RW */	
#define	REG_PAUSE			*(volatile u16*)0x4000300	/* ?? */

//---------------------------------------------------------------------
// Display Control Flags (in REG_DISPCNT)
//---------------------------------------------------------------------

#define	REG_DISPCNT_MODE_Mask				(0x0007)	/*  0x0000 - Video Modes Mask						*/
#define	REG_DISPCNT_MODE_0					(0x0 << 0)	/*           Video Mode 0 : 4 Text					*/
#define	REG_DISPCNT_MODE_1					(0x1 << 0)	/*           Video Mode 1 : 2 Text + 1 Rotated		*/
#define	REG_DISPCNT_MODE_2					(0x2 << 0)	/*           Video Mode 2 : 2 Rotated				*/
#define	REG_DISPCNT_MODE_3					(0x3 << 0)	/*           Video Mode 3 : Bitmap 240x160 16 bit	*/
#define	REG_DISPCNT_MODE_4					(0x4 << 0)	/*           Video Mode 4 : Bitmap ...				*/
#define	REG_DISPCNT_MODE_5					(0x5 << 0)	/*           Video Mode 5 : Bitmap ...				*/
#define	REG_DISPCNT_CGB_MODE  				(1 <<  4)	/*  0x0008 - GB/GB Color Legacy flag				*/
#define	REG_DISPCNT_BACKBUFFER				(1 <<  5)	/*  0x0010 - Active Buffer (Video Mode 4)			*/
#define	REG_DISPCNT_HBLANK_OAM_ACCESS		(1 <<  6)	/*  0x0020 - Allows OAM Accesses during H-Blanks	*/
#define	REG_DISPCNT_OBJ_MAP_2D				(0 <<  6)	/*  0x0040 - Sprite Tile Mapping : 2-D				*/
#define	REG_DISPCNT_OBJ_MAP_1D				(1 <<  6)	/*           Sprite Tile Mapping : 1-D				*/
#define	REG_DISPCNT_FORCED_BLANK			(1 <<  7)	/*  0x0080 - Forced Screen Blanking					*/
#define	REG_DISPCNT_BG0_ENABLE				(1 <<  8)	/*  0x0100 - Background 0 Enable					*/
#define	REG_DISPCNT_BG1_ENABLE				(1 <<  9)	/*  0x0200 - Background 1 Enable					*/
#define	REG_DISPCNT_BG2_ENABLE				(1 << 10)	/*  0x0400 - Background 2 Enable					*/
#define	REG_DISPCNT_BG3_ENABLE				(1 << 11)	/*  0x0800 - Background 3 Enable					*/
#define	REG_DISPCNT_OBJ_ENABLE				(1 << 12)	/*  0x1000 - Sprites Enable							*/
#define	REG_DISPCNT_WIN0_ENABLE				(1 << 13)	/*  0x2000 - Window 0 Enable						*/
#define	REG_DISPCNT_WIN1_ENABLE				(1 << 14)	/*  0x4000 - Window 1 Enable						*/
#define	REG_DISPCNT_WINOBJ_ENABLE			(1 << 15)	/*  0x8000 - Sprite Window Enable					*/
#define REG_DISPCNT_GREEN_SWAP				(1 << 16)	/* 0x10000 - Odd/even Green Data Swap				*/

//---------------------------------------------------------------------
// Display Stat	Flags (in REG_DISPSTAT)
//---------------------------------------------------------------------

#define	REG_DISPSTAT_VBLANK_IF				(1 << 0)	/*  0x0001 - Set when inside V-Blank interval		*/
#define	REG_DISPSTAT_HBLANK_IF				(1 << 1)	/*  0x0002 - Set when inside H-Blank interval		*/
#define	REG_DISPSTAT_VCOUNT_IF				(1 << 2)	/*  0x0004 - Set when V-Counter match given one		*/
#define	REG_DISPSTAT_VBLANK_IE				(1 << 3)	/*  0x0008 - V-Blank Interrupt Enable				*/
#define	REG_DISPSTAT_HBLANK_IE				(1 << 4)	/*  0x0010 - H-Blank Interrupt Enable				*/
#define	REG_DISPSTAT_VCOUNT_IE				(1 << 5)	/*  0x0020 - V-Counter Interrupt Enable				*/
#define	REG_DISPSTAT_VCOUNT_Mask			(0xFF00)	/*  0xFF00 - V-Counter Mask							*/

#define	REG_DISPSTAT_VCOUNT_Set(val)		(((val) << 8) GBALIB_MASK(REG_DISPSTAT_VCOUNT_Mask))

//---------------------------------------------------------------------
// Background Control Flags	(in	REG_BGxCNT)
//---------------------------------------------------------------------

#define	REG_BGxCNT_PRIORITY_Mask			(0x0003)	/*  0x0003 - Priority Mask							*/
#define	REG_BGxCNT_CHAR_BASE_Mask			(0x000C)	/*  0x000C - Character Base Block Mask				*/
#define	REG_BGxCNT_MOSAIC					(1 << 6)	/*	0x0040 - Mosaic Enable							*/
#define	REG_BGxCNT_COLOR_16					(0 << 7)	/*	0x0080 - Color Mode : 16 Colors					*/
#define	REG_BGxCNT_COLOR_256				(1 << 7)	/*           Color Mode : 256 Colors				*/
#define	REG_BGxCNT_SCREEN_BASE_Mask			(0x1F00)	/*  0x1F00 - Screen Base Block Mask					*/
#define	REG_BGxCNT_AREA_OVERFLOW_TRANS		(0 << 13)	/*	0x2000 - Area Overflow Flag : Transparent		*/
#define	REG_BGxCNT_AREA_OVERFLOW_WRAP		(1 << 13)	/*           Area Overflow Flag : Wrap				*/
#define	REG_BGxCNT_SCREEN_SIZE_Mask			(0xC000)	/*  0xC000 - Screen Size Mask						*/
#define	REG_BGxCNT_SCREEN_SIZE_256_256		(0x0 << 14)	/*           Screen Size : 256x256					*/
#define	REG_BGxCNT_SCREEN_SIZE_512_256		(0x1 << 14) /*           Screen Size : 512x256					*/
#define	REG_BGxCNT_SCREEN_SIZE_256_512		(0x2 << 14) /*           Screen Size : 256x512					*/
#define	REG_BGxCNT_SCREEN_SIZE_512_512		(0x3 << 14) /*           Screen Size : 512x512					*/

#define	REG_BGxCNT_PRIORITY_Set(val)		(((val) << 0) GBALIB_MASK(REG_BGxCNT_PRIORITY_Mask))
#define	REG_BGxCNT_CHAR_BASE_Set(val)		(((val) << 2) GBALIB_MASK(REG_BGxCNT_CHAR_BASE_Mask))
#define	REG_BGxCNT_CHAR_BASE_Get(reg)		(((reg) & REG_BGxCNT_CHAR_BASE_Mask) >> 2)
#define	REG_BGxCNT_CHAR_BASE_Addr(val)		((u16 *)((u8 *)MEM_VRAM_BG + (0x4000 * val)))
#define	REG_BGxCNT_SCREEN_BASE_Set(val)		(((val) << 8) GBALIB_MASK(REG_BGxCNT_SCREEN_BASE_Mask))
#define	REG_BGxCNT_SCREEN_BASE_Get(reg)		(((reg) & REG_BGxCNT_SCREEN_BASE_Mask) >> 8)
#define	REG_BGxCNT_SCREEN_BASE_Addr(val)	((u16 *)((u8 *)MEM_VRAM_BG + (0x0800 * val)))

//---------------------------------------------------------------------
// Windows Positionning	(in	REG_WIN[01][HV])
// see window.h	for	mappable structures	definitions
//---------------------------------------------------------------------

#define	REG_WIN0H_X2_Mask					(0x00FF)	/*  0x00FF - Window 0 X2 Mask						*/
#define	REG_WIN0H_X1_Mask					(0xFF00)	/*  0xFF00 - Window 0 X1 Mask						*/
#define	REG_WIN0V_Y2_Mask					(0x00FF)	/*  0x00FF - Window 0 Y2 Mask						*/
#define	REG_WIN0V_Y1_Mask					(0xFF00)	/*  0xFF00 - Window 0 Y1 Mask						*/
#define	REG_WIN1H_X2_Mask					(0x00FF)	/*  0x00FF - Window 1 X2 Mask						*/
#define	REG_WIN1H_X1_Mask					(0xFF00)	/*  0xFF00 - Window 1 X1 Mask						*/
#define	REG_WIN1V_Y2_Mask					(0x00FF)	/*  0x00FF - Window 1 Y2 Mask						*/
#define	REG_WIN1V_Y1_Mask					(0xFF00)	/*  0xFF00 - Window 1 Y1 Mask						*/

#define	REG_WIN0H_X2_Set(val)				(((val) << 0) GBALIB_MASK(REG_WIN0H_X2_Mask))
#define	REG_WIN0H_X1_Set(val)				(((val) << 8) GBALIB_MASK(REG_WIN0H_X1_Mask))
#define	REG_WIN0V_Y2_Set(val)				(((val) << 0) GBALIB_MASK(REG_WIN0V_Y2_Mask))
#define	REG_WIN0V_Y1_Set(val)				(((val) << 8) GBALIB_MASK(REG_WIN0V_Y1_Mask))
#define	REG_WIN1H_X2_Set(val)				(((val) << 0) GBALIB_MASK(REG_WIN1H_X2_Mask))
#define	REG_WIN1H_X1_Set(val)				(((val) << 8) GBALIB_MASK(REG_WIN1H_X1_Mask))
#define	REG_WIN1V_Y2_Set(val)				(((val) << 0) GBALIB_MASK(REG_WIN1V_Y2_Mask))
#define	REG_WIN1V_Y1_Set(val)				(((val) << 8) GBALIB_MASK(REG_WIN1V_Y1_Mask))

//---------------------------------------------------------------------
// Windows Inside Flags	(in	REG_WININ)
// see window.h	for	mappable structures	definitions
//---------------------------------------------------------------------

#define	REG_WININ_DISP_NONE					(0x0000)
#define	REG_WININ_DISP_BG0					(1 << 0)
#define	REG_WININ_DISP_BG1					(1 << 1)
#define	REG_WININ_DISP_BG2					(1 << 2)
#define	REG_WININ_DISP_BG3					(1 << 3)
#define	REG_WININ_DISP_OBJ					(1 << 4)
#define	REG_WININ_DISP_ALL					(0x001F)
#define	REG_WININ_COLOR_EFFECTS				(1 << 5)

#define	REG_WININ_WIN0_DISP_NONE			(REG_WININ_DISP_NONE		<< 0)
#define	REG_WININ_WIN0_DISP_BG0				(REG_WININ_DISP_BG0			<< 0)
#define	REG_WININ_WIN0_DISP_BG1				(REG_WININ_DISP_BG1			<< 0)
#define	REG_WININ_WIN0_DISP_BG2				(REG_WININ_DISP_BG2			<< 0)
#define	REG_WININ_WIN0_DISP_BG3				(REG_WININ_DISP_BG3			<< 0)
#define	REG_WININ_WIN0_DISP_OBJ				(REG_WININ_DISP_OBJ			<< 0)
#define	REG_WININ_WIN0_COLOR_EFFECTS		(REG_WININ_COLOR_EFFECTS	<< 0)
#define	REG_WININ_WIN1_DISP_NONE			(REG_WININ_DISP_NONE		<< 8)
#define	REG_WININ_WIN1_DISP_BG0				(REG_WININ_DISP_BG0			<< 8)
#define	REG_WININ_WIN1_DISP_BG1				(REG_WININ_DISP_BG1			<< 8)
#define	REG_WININ_WIN1_DISP_BG2				(REG_WININ_DISP_BG2			<< 8)
#define	REG_WININ_WIN1_DISP_BG3				(REG_WININ_DISP_BG3			<< 8)
#define	REG_WININ_WIN1_DISP_OBJ				(REG_WININ_DISP_OBJ			<< 8)
#define	REG_WININ_WIN1_COLOR_EFFECTS		(REG_WININ_COLOR_EFFECTS	<< 8)

//---------------------------------------------------------------------
// Windows Outside Flags (in REG_WINOUT)
// see window.h	for	mappable structures	definitions
//---------------------------------------------------------------------

#define	REG_WINOUT_DISP_NONE				(0x0000)
#define	REG_WINOUT_DISP_BG0					(1 << 0)
#define	REG_WINOUT_DISP_BG1					(1 << 1)
#define	REG_WINOUT_DISP_BG2					(1 << 2)
#define	REG_WINOUT_DISP_BG3					(1 << 3)
#define	REG_WINOUT_DISP_OBJ					(1 << 4)
#define	REG_WINOUT_DISP_ALL					(0x001F)
#define	REG_WINOUT_COLOR_EFFECTS			(1 << 5)

#define	REG_WINOUT_WIN01_DISP_NONE			(REG_WINOUT_DISP_NONE		<< 0)
#define	REG_WINOUT_WIN01_DISP_BG0			(REG_WINOUT_DISP_BG0		<< 0)
#define	REG_WINOUT_WIN01_DISP_BG1			(REG_WINOUT_DISP_BG1		<< 0)
#define	REG_WINOUT_WIN01_DISP_BG2			(REG_WINOUT_DISP_BG2		<< 0)
#define	REG_WINOUT_WIN01_DISP_BG3			(REG_WINOUT_DISP_BG3		<< 0)
#define	REG_WINOUT_WIN01_DISP_OBJ			(REG_WINOUT_DISP_OBJ		<< 0)
#define	REG_WINOUT_WIN01_DISP_ALL			(REG_WINOUT_DISP_ALL		<< 0)
#define	REG_WINOUT_WIN01_COLOR_EFFECTS		(REG_WINOUT_COLOR_EFFECTS	<< 0)
#define	REG_WINOUT_WINOBJ_DISP_NONE			(REG_WINOUT_DISP_NONE		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_BG0			(REG_WINOUT_DISP_BG0		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_BG1			(REG_WINOUT_DISP_BG1		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_BG2			(REG_WINOUT_DISP_BG2		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_BG3			(REG_WINOUT_DISP_BG3		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_OBJ			(REG_WINOUT_DISP_OBJ		<< 8)
#define	REG_WINOUT_WINOBJ_DISP_ALL			(REG_WINOUT_DISP_ALL		<< 8)
#define	REG_WINOUT_WINOBJ_COLOR_EFFECTS		(REG_WINOUT_COLOR_EFFECTS	<< 8)

//---------------------------------------------------------------------
// Mosaic Flags	(in	REG_MOSAIC)
//---------------------------------------------------------------------

#define	REG_MOSAIC_BG_H_SIZE_Mask			(0x000F)	/*  0x000F - Background Mosaic Horizontal Size Mask	*/
#define	REG_MOSAIC_BG_V_SIZE_Mask			(0x00F0)	/*  0x00F0 - Background Mosaic Vertical Size Mask	*/
#define	REG_MOSAIC_OBJ_H_SIZE_Mask			(0x0F00)	/*  0x0F00 - Background Mosaic Horizontal Size Mask	*/
#define	REG_MOSAIC_OBJ_V_SIZE_Mask			(0xF000)	/*	0xF000 - Background Mosaic Vertical Size Mask	*/

#define	REG_MOSAIC_BG_H_SIZE_Set(val)		(((val) <<  0) GBALIB_MASK(REG_MOSAIC_BG_H_SIZE_Mask))
#define	REG_MOSAIC_BG_V_SIZE_Set(val)		(((val) <<  4) GBALIB_MASK(REG_MOSAIC_BG_V_SIZE_Mask))
#define	REG_MOSAIC_BG_SIZE_Set(val)			(REG_MOSAIC_BG_H_SIZE_Set(val) | REG_MOSAIC_BG_V_SIZE_Set(val))
#define	REG_MOSAIC_OBJ_H_SIZE_Set(val)		(((val) <<  8) GBALIB_MASK(REG_MOSAIC_OBJ_H_SIZE_Mask))
#define	REG_MOSAIC_OBJ_V_SIZE_Set(val)		(((val) << 12) GBALIB_MASK(REG_MOSAIC_OBJ_V_SIZE_Mask))
#define	REG_MOSAIC_OBJ_SIZE_Set(val)		(REG_MOSAIC_OBJ_H_SIZE_Set(val) | REG_MOSAIC_OBJ_V_SIZE_Set(val))

//---------------------------------------------------------------------
// Blending	Control	Flags (in REG_CNT)
//---------------------------------------------------------------------

#define	REG_BLDCNT_TARGET1_BG0				(1 << 0)	/*  0x0001 - Target 1 Background 0 Enable			*/
#define	REG_BLDCNT_TARGET1_BG1				(1 << 1)	/*  0x0002 - Target 1 Background 1 Enable			*/
#define	REG_BLDCNT_TARGET1_BG2				(1 << 2)	/*  0x0004 - Target 1 Background 2 Enable			*/
#define	REG_BLDCNT_TARGET1_BG3				(1 << 3)	/*  0x0008 - Target 1 Background 3 Enable			*/
#define	REG_BLDCNT_TARGET1_OBJ				(1 << 4)	/*  0x0010 - Target 1 Object Enable					*/
#define	REG_BLDCNT_TARGET1_BACKDROP			(1 << 5)	/*  0x0020 - Target 1 Backdrop Enable				*/

#define	REG_BLDCNT_EFFECT_NONE				(0x0 << 6)	/*  0x00C0 - Effect Selection : None				*/
#define	REG_BLDCNT_EFFECT_ALPHA				(0x1 << 6)	/*           Effect Selection : Alpha				*/
#define	REG_BLDCNT_EFFECT_BRIGHTNESS_INC	(0x2 << 6)	/*           Effect Selection : Brightness Increase	*/
#define	REG_BLDCNT_EFFECT_BRIGHTNESS_DEC	(0x3 << 6)	/*           Effect Selection : Brightness Decrease */

#define	REG_BLDCNT_TARGET2_BG0				(1 <<  8)	/*  0x0100 - Target 2 Background 0 Enable			*/
#define	REG_BLDCNT_TARGET2_BG1				(1 <<  9)	/*  0x0200 - Target 2 Background 1 Enable			*/
#define	REG_BLDCNT_TARGET2_BG2				(1 << 10)	/*  0x0400 - Target 2 Background 2 Enable			*/
#define	REG_BLDCNT_TARGET2_BG3				(1 << 11)	/*  0x0800 - Target 2 Background 3 Enable			*/
#define	REG_BLDCNT_TARGET2_OBJ				(1 << 12)	/*  0x1000 - Target 2 Object Enable					*/
#define	REG_BLDCNT_TARGET2_BACKDROP			(1 << 13)	/*  0x2000 - Target 2 Backdrop Enable				*/

//---------------------------------------------------------------------
// Blending	Coefficients (in REG_BLDALPHA and REG_BLDY)
//---------------------------------------------------------------------

#define	REG_BLDALPHA_EVA_Mask				(0x001F)	/*  0x001F - EVA Mask								*/
#define	REG_BLDALPHA_EVB_Mask				(0x1F00)	/*  0x1F00 - EVB Mask								*/
#define	REG_BLDALPHA_EVA_Set(val)			(((val) << 0) GBALIB_MASK(REG_BLDALPHA_EVA_Mask))
#define	REG_BLDALPHA_EVB_Set(val)			(((val) << 8) GBALIB_MASK(REG_BLDALPHA_EVB_Mask))

#define	REG_BLDY_EVY_Mask					(0x001F)	/*  0x001F - EVY Mask								*/
#define	REG_BLDY_EVY_Set(val)				(((val) << 0) GBALIB_MASK(REG_BLDY_EVY_Mask))

//---------------------------------------------------------------------
// Sound Flags
//---------------------------------------------------------------------

#define REG_SNDCNT_L_SNDx_VOLUME_R_Mask		(0x0007)
#define REG_SNDCNT_L_SNDx_VOLUME_R_Set(val)	((val) << 0)
#define REG_SNDCNT_L_SNDx_VOLUME_L_Mask		(0x0070)
#define REG_SNDCNT_L_SNDx_VOLUME_L_Set(val)	((val) << 4)
#define REG_SNDCNT_L_SND1_ENABLE_R			(0x0100)
#define REG_SNDCNT_L_SND2_ENABLE_R			(0x0200)
#define REG_SNDCNT_L_SND3_ENABLE_R			(0x0400)
#define REG_SNDCNT_L_SND4_ENABLE_R			(0x0800)
#define REG_SNDCNT_L_SND1_ENABLE_L			(0x1000)
#define REG_SNDCNT_L_SND2_ENABLE_L			(0x2000)
#define REG_SNDCNT_L_SND3_ENABLE_L			(0x4000)
#define REG_SNDCNT_L_SND4_ENABLE_L			(0x8000)
#define REG_SNDCNT_L_SND1_ENABLE			(REG_SNDCNT_L_SND1_ENABLE_R | REG_SNDCNT_L_SND1_ENABLE_L)
#define REG_SNDCNT_L_SND2_ENABLE			(REG_SNDCNT_L_SND2_ENABLE_R | REG_SNDCNT_L_SND2_ENABLE_L)
#define REG_SNDCNT_L_SND3_ENABLE			(REG_SNDCNT_L_SND3_ENABLE_R | REG_SNDCNT_L_SND3_ENABLE_L)
#define REG_SNDCNT_L_SND4_ENABLE			(REG_SNDCNT_L_SND4_ENABLE_R | REG_SNDCNT_L_SND4_ENABLE_L)

#define REG_SNDCNT_H_SNDx_VOLUME_25			(0x0000)
#define REG_SNDCNT_H_SNDx_VOLUME_50			(0x0001)
#define REG_SNDCNT_H_SNDx_VOLUME_100		(0x0002)
#define REG_SNDCNT_H_DMA1_VOLUME_50			(0x0000)
#define REG_SNDCNT_H_DMA1_VOLUME_100		(0x0004)
#define REG_SNDCNT_H_DMA2_VOLUME_50			(0x0000)
#define REG_SNDCNT_H_DMA2_VOLUME_100		(0x0008)
#define REG_SNDCNT_H_DMA1_ENABLE_R			(0x0100)
#define REG_SNDCNT_H_DMA1_ENABLE_L			(0x0200)
#define REG_SNDCNT_H_DMA1_ENABLE			(REG_SNDCNT_H_DMA1_ENABLE_R | REG_SNDCNT_H_DMA1_ENABLE_L)
#define REG_SNDCNT_H_DMA1_TIMER_0			(0x0000)
#define REG_SNDCNT_H_DMA1_TIMER_1			(0x0400)
#define REG_SNDCNT_H_DMA1_RESET_FIFO		(0x0800)
#define REG_SNDCNT_H_DMA2_ENABLE_R			(0x1000)
#define REG_SNDCNT_H_DMA2_ENABLE_L			(0x2000)
#define REG_SNDCNT_H_DMA2_ENABLE			(REG_SNDCNT_H_DMA2_ENABLE_R | REG_SNDCNT_H_DMA2_ENABLE_L)
#define REG_SNDCNT_H_DMA2_TIMER_0			(0x0000)
#define REG_SNDCNT_H_DMA2_TIMER_1			(0x4000)
#define REG_SNDCNT_H_DMA2_RESET_FIFO		(0x8000)

#define REG_SNDCNT_X_SND1_ON				(0x0001)	/* Read Only */
#define REG_SNDCNT_X_SND2_ON				(0x0002)	/* Read Only */
#define REG_SNDCNT_X_SND3_ON				(0x0004)	/* Read Only */
#define REG_SNDCNT_X_SND4_ON				(0x0008)	/* Read Only */
#define REG_SNDCNT_X_ALL_SOUND_DISABLED		(0x0000)
#define REG_SNDCNT_X_ALL_SOUND_ENABLED		(0x0080)

#define REG_SNDBIAS_BIAS_LEVEL_Mask			(0x03FF)
#define REG_SNDBIAS_BIAS_LEVEL_Set(val)		((val) << 0)
#define REG_SNDBIAS_ARSC_9BITS_32KHZ		(0x0000)
#define REG_SNDBIAS_ARSC_8BITS_64KHZ		(0x4000)
#define REG_SNDBIAS_ARSC_7BITS_128KHZ		(0x8000)
#define REG_SNDBIAS_ARSC_6BITS_256KHZ		(0xC000)

//---------------------------------------------------------------------
// DMA Flags (in REG_DMAx_xxxx)
//---------------------------------------------------------------------
// See dma.h
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// Timer Control Flags (in REG_TMxCNT)
//---------------------------------------------------------------------

#define	REG_TMxCNT_H_PRESCALAR_SYSTEM		(0x0000)	/* 59.595 ns */
#define	REG_TMxCNT_H_PRESCALAR_64			(0x0001)	/*	3.814 us */
#define	REG_TMxCNT_H_PRESCALAR_256			(0x0002)	/* 15.256 us */
#define	REG_TMxCNT_H_PRESCALAR_1024			(0x0003)	/* 61.025 us */
#define	REG_TMxCNT_H_COUNTUP				(0x0004)	/* Timer 1/2/3 only */
#define	REG_TMxCNT_H_IE						(0x0040)
#define	REG_TMxCNT_H_ENABLE					(0x0080)

//---------------------------------------------------------------------
// Communication Flags
//---------------------------------------------------------------------

#define REG_RCNT_NORMAL_MULTI_UART_MODE		(0x0000)
#define REG_RCNT_GENERAL_PURPOSE_MODE		(0x8000)
#define REG_RCNT_JOYBUS_MODE				(0xC000)

#define REG_RCNT_SC							(0x0001)
#define REG_RCNT_SD							(0x0002)
#define REG_RCNT_SI							(0x0004)
#define REG_RCNT_SO							(0x0008)
#define REG_RCNT_SC_DIRECTION				(0x0010)
#define REG_RCNT_SD_DIRECTION				(0x0020)
#define REG_RCNT_SI_DIRECTION				(0x0040)
#define REG_RCNT_S0_DIRECTION				(0x0080)
#define REG_RCNT_IE							(0x0100)

#define REG_SIOCNT_START					(0x0080)
#define REG_SIOCNT_NORMAL_MODE				(0x0000)
#define REG_SIOCNT_MULTI_PLAYER_MODE		(0x2000)
#define REG_SIOCNT_UART_MODE				(0x3000)
#define REG_SIOCNT_IE						(0x4000)

// In Normal Mode
#define REG_SIOCNT_SC_EXTERNAL				(0x0000)
#define REG_SIOCNT_SC_INTERNAL				(0x0001)
#define REG_SIOCNT_SC_INTERNAL_256KHZ		(0x0000)
#define REG_SIOCNT_SC_INTERNAL_2048KHZ		(0x0002)
#define REG_SIOCNT_SI_STATE_LOW				(0x0000)	/* Read Only */
#define REG_SIOCNT_SI_STATE_HIGH_NONE		(0x0004)	/* Read Only */
#define REG_SIOCNT_SO_INACTIVITY_LOW		(0x0000)
#define REG_SIOCNT_SO_INACTIVITY_HIGH		(0x0008)
#define REG_SIOCNT_TRANSFER_LENGTH_8BITS	(0x0000)
#define REG_SIOCNT_TRANSFER_LENGTH_32BITS	(0x1000)

// In Multi-Player Mode
#define REG_SIOCNT_BAUD_RATE_9600			(0x0000)
#define REG_SIOCNT_BAUD_RATE_38400			(0x0001)
#define REG_SIOCNT_BAUD_RATE_57600			(0x0002)
#define REG_SIOCNT_BAUD_RATE_115200			(0x0003)
#define REG_SIOCNT_SI_TERMINAL_PARENT		(0x0000)	/* Read Only */
#define REG_SIOCNT_SI_TERMINAL_CHILD		(0x0004)	/* Read Only */
#define REG_SIOCNT_SD_TERMINAL_BAD_CONN		(0x0000)	/* Read Only */
#define REG_SIOCNT_SD_TERMINAL_ALL_READY	(0x0008)	/* Read Only */
#define REG_SIOCNT_MULTI_PLAYER_ID_Mask		(0x0030)	/* Read Only */
#define REG_SIOCNT_MULTI_PLAYER_ID_Set(val)	((val) << 4)
#define REG_SIOCNT_ERROR					(0x0040)

// In UART Mode
// Baud Rate: as in Multi-Player Mode
#define REG_SIOCNT_UART_CTS_SEND_ALWAYS		(0x0000)
#define REG_SIOCNT_UART_CTS_SEND_ON_SC_LOW	(0x0004)
#define REG_SIOCNT_UART_PARITY				(0x0008)
#define REG_SIOCNT_UART_SEND_DATA_FULL		(0x0010)	/* Read Only */
#define REG_SIOCNT_UART_RECEIVE_DATA_EMPTY	(0x0020)	/* Read Only */
// Error: as in Multi-Player Mode
#define REG_SIOCNT_UART_DATA_LENGTH_7BITS	(0x0000)
#define REG_SIOCNT_UART_DATA_LENGTH_8BITS	(0x0080)
#define REG_SIOCNT_UART_FIFO_ENABLE			(0x0100)
#define REG_SIOCNT_UART_PARITY_ENABLE		(0x0200)
#define REG_SIOCNT_UART_SEND_ENABLE			(0x0400)
#define REG_SIOCNT_UART_RECEIVE_ENABLE		(0x0800)

//---------------------------------------------------------------------
// Communication / Inputs Flags (JOY_BUS mode)
//---------------------------------------------------------------------

#define REG_JOYCNT_DEVICE_RESET				(0x0001)	/* Read / Ack */
#define REG_JOYCNT_RECEIVE_COMPLETE			(0x0002)	/* Read / Ack */
#define REG_JOYCNT_SEND_COMPLETE			(0x0004)	/* Read / Ack */
#define REG_JOYCNT_DEVICE_RESET_IE			(0x0040)

#define REG_JOYSTAT_RECEIVE_IF				(0x0002)	/* Read Only ? */
#define REG_JOYSTAT_SEND_IF					(0x0008)	/* Read Only ? */
#define REG_JOYSTAT_USER_FLAGS_Mask			(0x0030)
#define REG_JOYSTAT_USER_FLAGS_Set(val)		((val) << 4)

//---------------------------------------------------------------------
// Key Inputs Flags	(in	REG_KEYINPUT)
//---------------------------------------------------------------------

#define	REG_KEYINPUT_KEY_A					(1 << 0)	/*  0x0001 - Key : A								*/
#define	REG_KEYINPUT_KEY_B					(1 << 1)	/*  0x0002 - Key : B								*/
#define	REG_KEYINPUT_KEY_SELECT				(1 << 2)	/*  0x0004 - Key : Select							*/
#define	REG_KEYINPUT_KEY_START				(1 << 3)	/*  0x0008 - Key : Start							*/
#define	REG_KEYINPUT_KEY_RIGHT				(1 << 4)	/*  0x0010 - Key : Right							*/
#define	REG_KEYINPUT_KEY_LEFT				(1 << 5)	/*  0x0020 - Key : Left								*/
#define	REG_KEYINPUT_KEY_UP					(1 << 6)	/*  0x0040 - Key : Up								*/
#define	REG_KEYINPUT_KEY_DOWN				(1 << 7)	/*  0x0080 - Key : Down								*/
#define	REG_KEYINPUT_KEY_R					(1 << 8)	/*  0x0100 - Key : R								*/
#define	REG_KEYINPUT_KEY_L					(1 << 9)	/*  0x0200 - Key : L								*/
#define REG_KEYINPUT_KEYS_RESET				(REG_KEYINPUT_KEY_A | REG_KEYINPUT_KEY_B | REG_KEYINPUT_KEY_SELECT | REG_KEYINPUT_KEY_START)

//---------------------------------------------------------------------
// Key Control Flags (in REG_KEYCNT)
//---------------------------------------------------------------------

#define	REG_KEYCNT_KEY_A					(1 <<  0)	/*  0x0001 - Key : A								*/
#define	REG_KEYCNT_KEY_B					(1 <<  1)	/*  0x0002 - Key : B								*/
#define	REG_KEYCNT_KEY_SELECT				(1 <<  2)	/*  0x0004 - Key : Select							*/
#define	REG_KEYCNT_KEY_START				(1 <<  3)	/*  0x0008 - Key : Start							*/
#define	REG_KEYCNT_KEY_RIGHT				(1 <<  4)	/*  0x0010 - Key : Right							*/
#define	REG_KEYCNT_KEY_LEFT					(1 <<  5)	/*  0x0020 - Key : Left								*/
#define	REG_KEYCNT_KEY_UP					(1 <<  6)	/*  0x0040 - Key : Up								*/
#define	REG_KEYCNT_KEY_DOWN					(1 <<  7)	/*  0x0080 - Key : Down								*/
#define	REG_KEYCNT_KEY_R					(1 <<  8)	/*  0x0100 - Key : R								*/
#define	REG_KEYCNT_KEY_L					(1 <<  9)	/*  0x0200 - Key : L								*/
#define	REG_KEYCNT_IE						(1 << 14)	/*  0x4000 - Interrupt Enable						*/
#define	REG_KEYCNT_INT_CONDITION_OR			(0 << 15)	/*  0x8000 - Interrupt Condition : Or (any key)		*/
#define	REG_KEYCNT_INT_CONDITION_AND		(1 << 15)	/*           Interrupt Condition : And (all keys)	*/

//---------------------------------------------------------------------
// Interrupt Flags (in REG_IE &	REG_IME)
//---------------------------------------------------------------------

#define	REG_IE_NONE							(0x0000)
#define	REG_IE_VBLANK						(1 <<  0)	/*  0x0001 - V-Blank Interrupt Enable				*/
#define	REG_IE_HBLANK						(1 <<  1)	/*  0x0002 - H-Blank Interrupt Enable				*/
#define	REG_IE_VCOUNT						(1 <<  2)	/*  0x0004 - V-Count Interrupt Enable				*/
#define	REG_IE_TM0							(1 <<  3)	/*  0x0008 - Timer 0 Interrupt Enable				*/
#define	REG_IE_TM1							(1 <<  4)	/*  0x0010 - Timer 1 Interrupt Enable				*/
#define	REG_IE_TM2							(1 <<  5)	/*  0x0020 - Timer 2 Interrupt Enable				*/
#define	REG_IE_TM3							(1 <<  6)	/*  0x0040 - Timer 3 Interrupt Enable				*/
#define	REG_IE_COMM							(1 <<  7)	/*  0x0080 - Communication Interrupt Enable			*/
#define	REG_IE_DMA0							(1 <<  8)	/*  0x0100 - DMA 0 Interrupt Enable					*/
#define	REG_IE_DMA1							(1 <<  9)	/*  0x0200 - DMA 1 Interrupt Enable					*/
#define	REG_IE_DMA2							(1 << 10)	/*  0x0400 - DMA 2 Interrupt Enable					*/
#define	REG_IE_DMA3							(1 << 11)	/*  0x0800 - DMA 3 Interrupt Enable					*/
#define	REG_IE_KEY							(1 << 12)	/*  0x1000 - Key Interrupt Enable					*/
#define	REG_IE_GAMEPAK						(1 << 13)	/*  0x2000 - GamePak Interrupt Enable				*/
#define	REG_IE_ALL							(0x3FFF)

#define	REG_IME_DISABLE						(0 << 0)	/*  0x0001 - Master Interrupt : Disable				*/
#define	REG_IME_ENABLE						(1 << 0)	/*           Master Interrupt : Enable				*/

//---------------------------------------------------------------------
// BACKGROUND MAPPING
//---------------------------------------------------------------------

// Text BG
#define	BG_MAP_CHAR_Mask					(0x03FF) 	/*  0x03FF - Character Index Mask					*/
#define	BG_MAP_FLIP_NONE					(0 << 10)	/*           No Flipping							*/
#define	BG_MAP_FLIP_H						(1 << 10)	/*  0x0400 - Horizontal Flip Flag					*/
#define	BG_MAP_FLIP_V						(1 << 11)	/*  0x0800 - Vertical Flip Flag						*/
#define	BG_MAP_PALETTE_Mask					(0xF000)	/*	0xF000 - Palette (16-color mode index)			*/
#define	BG_MAP_CHAR_Set(val)				(((val) <<  0) GBALIB_MASK(BG_MAP_CHAR_Mask))
#define	BG_MAP_PALETTE_Set(val)				(((val) << 12) GBALIB_MASK(BG_MAP_PALETTE_Mask))
          
// Rotation/Scaling BG                              	
#define	BG_ROT_MAP_CHAR_Mask				(0x00FF)	/*  0x00FF - Character Index Mask 					*/
#define	BG_ROT_MAP_CHAR_Set(val)			(((val) <<  0) GBALIB_MASK(BG_ROT_MAP_CHAR_Mask))

//---------------------------------------------------------------------
// SPRITES
//---------------------------------------------------------------------
// see sprites.h for mappable structures definitions
//---------------------------------------------------------------------

#define	SPR_MAX								(128)
                                        	
// Separately accessed (using bitfields)	
                                        	
#define	SPR_MODE_NORMAL						(0)
#define	SPR_MODE_TRANS						(1)
#define	SPR_MODE_WINDOW						(2)
#define	SPR_COLOR_16						(0)
#define	SPR_COLOR_256						(1)
#define	SPR_SHAPE_SQUARE					(0)
#define	SPR_SHAPE_WIDE						(1)
#define	SPR_SHAPE_TALL						(2)
#define	SPR_SIZE_8							(0)
#define	SPR_SIZE_16							(1)
#define	SPR_SIZE_32							(2)
#define	SPR_SIZE_64							(3)

// In Attributes

#define	SPR_A0_Y_Mask						(0x00FF) /*	Y Coordinate */
#define	SPR_A0_Y_Set(val)					(((val) << 0) GBALIB_MASK(SPR_A0_Y_Mask))
#define	SPR_A0_ROTATE						(0x0100) /*	Rotate flag	 */
#define	SPR_A0_DOUBLE						(0x0200) /*	Double flag	 */
#define	SPR_A0_MODE_NORMAL					(0x0000) /*	Blend Mode	 */
#define	SPR_A0_MODE_TRANS					(0x0400)
#define	SPR_A0_MODE_WINDOW					(0x0800)
#define	SPR_A0_MOSAIC						(0x1000) /*	Mosaic flag	 */
#define	SPR_A0_COLOR_16						(0x0000) /*	Color flag	 */
#define	SPR_A0_COLOR_256					(0x2000)
#define	SPR_A0_SHAPE_Mask					(0xC000)
#define	SPR_A0_SHAPE_SQUARE					(0x0000) /*	Shape		 */
#define	SPR_A0_SHAPE_WIDE					(0x4000)
#define	SPR_A0_SHAPE_TALL					(0x8000)
                                        	
#define	SPR_A1_X_Mask						(0x01FF) /*	X Coordinate */
#define	SPR_A1_X_Set(val)					(((val) << 0) GBALIB_MASK(SPR_A1_X_Mask))
#define	SPR_A1_ROTIDX_Mask					(0x0E00) /*	Rot	index	 */
#define	SPR_A1_ROTIDX_Set(val)				(((val) << 9) GBALIB_MASK(SPR_A1_ROTIDX_Mask))
#define	SPR_A1_FLIP_NONE					(0x0000) /*	No Flipping	 */
#define	SPR_A1_FLIP_H						(0x1000) /*	H Flip flag	 */
#define	SPR_A1_FLIP_V						(0x2000) /*	V Flip flag	 */
#define	SPR_A1_SIZE_8						(0x0000) /*	Size		 */
#define	SPR_A1_SIZE_16						(0x4000)
#define	SPR_A1_SIZE_32						(0x8000)
#define	SPR_A1_SIZE_64						(0xC000)
                                        	
#define	SPR_A2_CHAR_Mask					(0x03FF) /*	Tile index	 */
#define	SPR_A2_CHAR_Set(val)				(((val) <<  0) GBALIB_MASK(SPR_A2_CHAR_Mask))
#define SPR_A2_PRIORITY_Mask				(0x0C00)
#define	SPR_A2_PRIORITY_Set(val)			(((val) << 10) GBALIB_MASK(SPR_A2_PRIORITY_Mask))
#define	SPR_A2_PRIORITY_0					(0x0000)
#define	SPR_A2_PRIORITY_1					(0x0400)
#define	SPR_A2_PRIORITY_2					(0x0800)
#define	SPR_A2_PRIORITY_3					(0x0C00)
#define	SPR_A2_PALETTE_Mask					(0xF000)
#define	SPR_A2_PALETTE_Set(val)				(((val) << 12) GBALIB_MASK(SPR_A2_PALETTE_Mask))

//---------------------------------------------------------------------
