//---------------------------------------------------------------------
// GBALIB: msprintf.c
// by Omar Cornut
//---------------------------------------------------------------------
// Minimal [v]sprintf() function
//---------------------------------------------------------------------
// Taken from the linux (2.4.0) sources
// Rewritten with features removed, in order to avoid linking LibC
// one which add a ~25/30 kb overhead of the program.
//---------------------------------------------------------------------

#include "gbalib.h"

#define isdigit(c)  (((c) >= '0') && ((c) <= '9'))

#define do_div(n,base)			\
({								\
        int __res;				\
        __res = n % base;		\
        n = n / base;			\
        __res;					\
})


static int skip_atoi(const char **s)
{
	int i = 0;

	while (isdigit(**s))
		i = (i * 10) + *((*s)++) - '0';
	return i;
}

#define ZEROPAD	1		/* pad with zero		*/
#define SIGN	2		/* unsigned/signed long	*/
#define PLUS	4		/* show plus			*/
#define SPACE	8		/* space if plus 		*/
#define LEFT	16		/* left justified 		*/
#define SPECIAL	32		/* 0x					*/
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */

static char * number(char * str, unsigned num, unsigned base,
                            int size, int precision, int type)
{
	char 		c, sign, tmp[66];
	const char *digits = "0123456789abcdefghijklmnopqrstuvwxyz";
	int 		i;

	if (type & LARGE)
		digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	if (type & LEFT)
		type &= ~ZEROPAD;
	if (base < 2 || base > 36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ';
	sign = 0;
	if (type & SIGN) 
	{
		if (num < 0) 
		{
			sign = '-';
			num = -num;
			size--;
		}
		else 
		if (type & PLUS) 
		{
			sign = '+';
			size--;
		} 
		else 
		if (type & SPACE) 
		{
			sign = ' ';
			size--;
		}
	}
	if (type & SPECIAL) 
	{
		if (base == 16)
			size -= 2;
		else 
		if (base == 8)
			size--;
	}
	i = 0;
	if (num == 0)
		tmp[i++]='0';
	else while (num != 0)
		tmp[i++] = digits[do_div(num,base)];
	if (i > precision)
		precision = i;
	size -= precision;
	if (!(type & (ZEROPAD+LEFT)))
		while (size-->0)
			*str++ = ' ';
	if (sign)
		*str++ = sign;
	if (type & SPECIAL) 
	{
		if (base==8)
			*str++ = '0';
		else if (base==16) 
		{
			*str++ = '0';
			*str++ = digits[33];
		}
	}
	if (!(type & LEFT))
		while (size-- > 0)
			*str++ = c;
	while (i < precision--)
		*str++ = '0';
	while (i-- > 0)
		*str++ = tmp[i];
	while (size-- > 0)
		*str++ = ' ';
	return str;
}

unsigned int 	strnlen(const char *s, int count)
{
	const char *sc;

	for (sc = s; count-- && *sc != '\0'; ++sc);
	return (sc - s);
}

// FUNCTION: mvsprintf()
// Print formatted string to given buffer, getting parameters from given stack pointer
int 			mvsprintf(char *buf, const char *format, va_list args)
{
	int			i;
	int			len;
	unsigned 	num, base;
	char 		*str;
	const char 	*s;

	int 		flags;			/* flags to number() */
	int 		field_width;	/* width of output field */
	int 		precision;		/* min. # of digits for integers; max number of chars for from string */
	int 		qualifier;		/* 'h', 'l', or 'L' for integer fields */
								/* 'z' support added 23/7/1999 S.H.    */
								/* 'z' changed to 'Z' --davidm 1/25/99 */
	
	for (str = buf; *format; ++format)
	{
		if (*format != '%') 
		{
			*str++ = *format;
			continue;
		}
			
		/* Process flags */
		flags = 0;
		flags_repeat:
			format++;	/* this also skips first '%' */
			switch (*format) 
			{
				case '-': flags |= LEFT; 	goto flags_repeat;
				case '+': flags |= PLUS; 	goto flags_repeat;
				case ' ': flags |= SPACE; 	goto flags_repeat;
				case '#': flags |= SPECIAL;	goto flags_repeat;
				case '0': flags |= ZEROPAD;	goto flags_repeat;
			}
		
		/* Get field width */
		field_width = -1;
		if (isdigit(*format))
			field_width = skip_atoi(&format);
		else 
		if (*format == '*') 
		{
			++format;
			/* It's the next argument */
			field_width = va_arg(args, int);
			if (field_width < 0) 
			{
				field_width = -field_width;
				flags |= LEFT;
			}
		}

		/* Get the precision */
		precision = -1;
		if (*format == '.') 
		{
			++format;
			if (isdigit(*format))
				precision = skip_atoi(&format);
			else if (*format == '*') 
			{
				++format;
				/* It's the next argument */
				precision = va_arg(args, int);
			}
			if (precision < 0)
				precision = 0;
		}

		/* Get the conversion qualifier */
		qualifier = -1;
		if (*format == 'h' || *format == 'l' || *format == 'L' || *format =='Z') 
			qualifier = *format++;

		/* Default base */
		base = 10;

		switch (*format) 
		{
		case 'c':
			if (!(flags & LEFT))
				while (--field_width > 0)
					*str++ = ' ';
			*str++ = (unsigned char) va_arg(args, int);
			while (--field_width > 0)
				*str++ = ' ';
			continue;

		case 's':
			s = va_arg(args, char *);
			if (!s)
				s = "<NULL>";

			len = strnlen(s, precision);

			if (!(flags & LEFT))
				while (len < field_width--)
					*str++ = ' ';
			for (i = 0; i < len; ++i)
				*str++ = *s++;
			while (len < field_width--)
				*str++ = ' ';
			continue;

		case 'p':
			if (field_width == -1) {
				field_width = 2*sizeof(void *);
				flags |= ZEROPAD;
			}
			str = number(str,
				(unsigned long) va_arg(args, void *), 16,
				field_width, precision, flags);
			continue;


		case 'n':
			if (qualifier == 'l') {
				long * ip = va_arg(args, long *);
				*ip = (str - buf);
			} else if (qualifier == 'Z') {
				int * ip = va_arg(args, int *);
				*ip = (str - buf);
			} else {
				int * ip = va_arg(args, int *);
				*ip = (str - buf);
			}
			continue;

		case '%':
			*str++ = '%';
			continue;

		/* Integer number formats - set up the flags and "break" */
		case 'o':
			base = 8;
			break;
		case 'X':
			flags |= LARGE;
		case 'x':
			base = 16;
			break;

		case 'd':
		case 'i':
			flags |= SIGN;
		case 'u':
			break;

		default:
			*str++ = '%';
			if (*format)
				*str++ = *format;
			else
				--format;
			continue;
		}
		
		if (qualifier == 'Z') {
			num = va_arg(args, int);
		} else if (qualifier == 'h') {
			num = (unsigned short) va_arg(args, int);
			if (flags & SIGN)
				num = (signed short) num;
		} else {
			num = va_arg(args, unsigned int);
			if (flags & SIGN)
				num = (signed int) num;
		}
		str = number(str, num, base, field_width, precision, flags);
	}
	
	*str = '\0';
	return (str - buf);
}

// FUNCTION: msprintf()
// Print formatted string to given buffer, getting parameters from stack
int			msprintf(char *buf, const char *format, ...)
{
	int 	i;
	va_list args;

	va_start(args, format);
	i = mvsprintf(buf, format, args);
	va_end(args);
	return (i);
}

