//---------------------------------------------------------------------
// GBALIB: sprites.c
// by Omar Cornut
//---------------------------------------------------------------------
// Sprites Handling System
//---------------------------------------------------------------------
// Two layers:
// - sprite[s]_oam
//   - Provide hardware mappable structures
//	   Handle low level shadowing and copying to OAM
// - sprite[s]
//   - Higher level sprite handling
//     Handle sprites sorting based on a user given sub_priority value
//     Handle sprites position offsets (used by the animation system)
//---------------------------------------------------------------------
 
#include "gbalib.h"

//---------------------------------------------------------------------
// SPRITES DATA
//---------------------------------------------------------------------

t_sprite		sprites[SPR_MAX+1];
t_sprite_oam	sprites_oam[SPR_MAX];
static int		sprites_copied;
static int		sprites_first_unused;
static int		sprites_oam_first_unused;

//---------------------------------------------------------------------
// OAM SPRITES FUNCTIONS
//---------------------------------------------------------------------

// FUNCTION: sprites_oam_copy()
// Copy shadowed OAM to hardware mapped OAM
inline void		sprites_oam_copy(void)
{
	DMA_Copy(3, DMA_32_NOW, (void *)sprites_oam, (void *)MEM_OAM, SPR_MAX * (sizeof (t_sprite_oam) / sizeof (u32)));
}

// FUNCTION: sprites_oam_clear()
// Clear shadowed OAM and hardware mapped OAM
inline void		sprites_oam_clear(void)
{
	int			i;

	for (i = 0; i < SPR_MAX; i++)
	{
		sprites_oam[i].attr0 = SPR_A0_Y_Set(RES_HEIGHT);
		sprites_oam[i].attr1 = SPR_A1_X_Set(RES_WIDTH);
	}
	sprites_oam_first_unused = 0;
	sprites_oam_copy();
}

// FUNCTION: sprite_oam_new()
// Return index of first unused sprite in OAM
// FIXME: try implementing _new/delete() using a stack of available sprites ?
int				sprite_oam_new(void)
{
	int			idx;

	// Find first unused cell	
	idx = sprites_oam_first_unused;
	while (++sprites_oam_first_unused < 128)
		if (sprites_oam[sprites_oam_first_unused].y == RES_HEIGHT)
			break;
	return (idx);
}

// FUNCTION: sprite_oam_delete()
// Set given sprite to unused state
void			sprite_oam_delete(int idx)
{
	sprites_oam[idx].y = RES_HEIGHT;
	sprites_oam[idx].x = RES_WIDTH;
	if (sprites_oam_first_unused > idx)
		idx = sprites_oam_first_unused;
}

// FUNCTION: sprite_oam_rotate()
// Setup rotation data for given index
void 			sprite_oam_rotate(int rotdata_idx, int angle)
{
	// Note: no need to read the COS() and SIN() once since they are table lookups anyway
	sprites_oam_rotdata[rotdata_idx].pa =  COS(angle);
	sprites_oam_rotdata[rotdata_idx].pb =  SIN(angle);
	sprites_oam_rotdata[rotdata_idx].pc = -SIN(angle);
	sprites_oam_rotdata[rotdata_idx].pd =  COS(angle);
}

// FUNCTION: sprite_oam_rotate_scale()
// Setup rotation & scaling data for given index
void 			sprite_oam_rotate_scale(int rotdata_idx, int angle, fu16 x_scale, fu16 y_scale)
{
	sprites_oam_rotdata[rotdata_idx].pa = FMUL(x_scale,  COS(angle), 8);
	sprites_oam_rotdata[rotdata_idx].pb = FMUL(y_scale,  SIN(angle), 8);
	sprites_oam_rotdata[rotdata_idx].pc = FMUL(x_scale, -SIN(angle), 8);
	sprites_oam_rotdata[rotdata_idx].pd = FMUL(y_scale,  COS(angle), 8);
}

//---------------------------------------------------------------------
// HIGH-LEVEL SPRITES FUNCTIONS
//---------------------------------------------------------------------

// FUNCTION: sprites_copy()
// Copy high-level sprites to hardware mapped OAM, passing thru shadowed OAM
void				sprites_copy(void)
{
	int				src_idx;		// in sprites[]
	t_sprite *		src_sprite;		// in sprites[]
	t_sprite_oam *	dst_sprite_oam;	// in sprites_oam[]

	// Create Shadow OAM from our sorted sprite list	
	src_idx = sprites[0].next;
	dst_sprite_oam = &sprites_oam[0];
	while (src_idx != 0)
	{
		src_sprite = &sprites[src_idx];
		if (src_sprite->offsets != 0)
		{
			// Handle X/Y offsets
			dst_sprite_oam->attr0 = (src_sprite->oam.attr0 & ~SPR_A0_Y_Mask) | ((src_sprite->oam.y + src_sprite->y_offset) & SPR_A0_Y_Mask);
			dst_sprite_oam->attr1 = (src_sprite->oam.attr1 & ~SPR_A1_X_Mask) | ((src_sprite->oam.x + src_sprite->x_offset) & SPR_A1_X_Mask);
		}
		else
		{
			// Direct copy
			dst_sprite_oam->attr0 = src_sprite->oam.attr0;
			dst_sprite_oam->attr1 = src_sprite->oam.attr1;
		}
		dst_sprite_oam->attr2 = src_sprite->oam.attr2;
		src_idx = src_sprite->next;
		dst_sprite_oam++;
	}

	// Clear now unused sprites in Shadow OAM, and set 'sprites_copied' for future call
	int left_to_clear = sprites_copied;
	sprites_copied = dst_sprite_oam - sprites_oam;
	left_to_clear -= sprites_copied;
	while (left_to_clear > 0)
	{
		dst_sprite_oam->attr0 = SPR_A0_Y_Set(RES_HEIGHT);
		dst_sprite_oam->attr1 = SPR_A1_X_Set(RES_WIDTH);
		left_to_clear--;
		dst_sprite_oam++;
	}

	// Copy Shadow OAM to OAM
	sprites_oam_copy();
}

// FUNCTION: sprites_clear()
// Clear OAM and high-level sprite list
void			sprites_clear(void)
{
	int			i;

	// Mark all sprites as unused
	for (i = 0; i < SPR_MAX + 1; i++)
		sprites[i].sub_priority = 0;

	// Clear sprites list
	sprites[0].next = 0;
	sprites_copied = SPR_MAX;
	sprites_first_unused = 1;
	sprites_copy();
}

#ifdef DEBUG_SPRITES
// [PRIVATE] FUNCTION: sprite_debug()
// Print list of allocated sprites
void			sprite_debug(void)
{
	int			idx;
	
	dprintf("[ sprite_debug\n[");
	idx = 0;
	do
	{
		dprintf("%0d: next->%d, ", idx, sprites[idx].next);
		idx = sprites[idx].next;
	}
	while (idx != 0);
	dprintf("]\n");
}
#endif

// FUNCTION: sprite_new()
// Return index of first unused sprite
// Must provide sprite sub_priority
int				sprite_new(int sub_priority)
{
	int			prev;
	int			next;

#ifdef DEBUG_SPRITES
	dprintf("sprite_new(%d)\n", sub_priority);
#endif
	
	// Find previous and next sprites
	// FIXME: optimize this search
	prev = 0;
	next = sprites[0].next;
	while (next != 0)
	{
		if (sprites[next].sub_priority <= sub_priority)
			break;
		prev = next;
		next = sprites[next].next;
	}
	//dprintf(" - prev: %d, next: %d\n", prev, next);
	
	// Find first unused cell
	int idx = sprites_first_unused;
	while (++sprites_first_unused < SPR_MAX + 1)
		if (sprites[sprites_first_unused].sub_priority == 0)
			break;
	
	// Fill cell and modify list
	sprites[prev].next = idx;
	sprites[idx].next = next;
	sprites[idx].sub_priority = sub_priority;
	sprites[idx].offsets = 0;

#ifdef DEBUG_SPRITES
	dprintf(" --> returning %d\n", idx);
#endif
	return (idx);
}

// FUNCTION: sprite_delete()
// Set given sprite to unused state
void			sprite_delete(int idx)
{	
#ifdef DEBUG_SPRITES
	dprintf("sprite_delete(%d), sprites_first_unused = %d\n", idx, sprites_first_unused);
	sprite_debug();
	dprintf("--\n");
#endif
	
	// Find previous sprites
	// FIXME: Optimize
	t_sprite *prev = &sprites[0];
	while (prev->next != idx)
		prev = &sprites[prev->next];

	// Remove sprite from list
	prev->next = sprites[idx].next;
	
	// Mark sprite as unused
	sprites[idx].sub_priority = 0;

	if (idx < sprites_first_unused)
		sprites_first_unused = idx;
}

// FUNCTION: sprite_priority_set()
// Change sub_priority of given sprite
void			sprite_priority_set(int idx, int sub_priority)
{

	// Setup new priority
	if (sprites[idx].sub_priority == sub_priority)
		return;
	sprites[idx].sub_priority = sub_priority;
	
	// Find previous & next cell in lists, for old and new positions
	u8 	prev = 0;
	u8 	next = sprites[0].next;
	u8	new_prev = 0;
	u8	new_next = 0;
	u8	old_prev = 0;
	u8	done = 0;
	while (next != 0)
	{
		if (next == idx)
		{
			old_prev = prev;
			if ((done |= 1) == 3)
				break;
		}
		else
		{
			if (((done & 2) == 0) && (sprites[next].sub_priority <= sub_priority))
			{
				new_prev = prev;
				new_next = next;
				if ((done |= 2) == 3)
		 			break;
			}
			prev = next;
		}
		next = sprites[next].next;
	}
	if (next == 0)
	{
		new_prev = prev;
		new_next = 0;
	}
	
	// Modify lists
	if (new_prev != old_prev)
	{
		sprites[old_prev].next = sprites[idx].next;
		sprites[new_prev].next = idx;
		sprites[idx].next = new_next;
	}
}

//---------------------------------------------------------------------
