//---------------------------------------------------------------------
// GBALIB: sprites.h
// by Omar Cornut
//---------------------------------------------------------------------
// Sprites Handling System
//---------------------------------------------------------------------
// Two layers:
// - sprite[s]_oam
//   - Provide hardware mappable structures
//	   Handle low level shadowing and copying to OAM
// - sprite[s]
//   - Higher level sprite handling
//     Handle sprites sorting based on a user given sub_priority value
//     Handle sprites position offsets (used by the animation system)
//---------------------------------------------------------------------

// Define DEBUG_SPRITES to enable sprites system core debugging
// Note: DEBUG must be defined as well
// #define DEBUG_SPRITES

//---------------------------------------------------------------------
// OAM SPRITES STRUCTURES
//---------------------------------------------------------------------

typedef struct
{
	union
	{
		struct
		{
								// 15    14       13          12         11       10       9            8          7-0
		  	u16	attr0; 			// shape(2)-  -256c(1)--  -mosaic(1)-  ---mode(2)---- -double(1)-  --rotate(1)--  -y(8)-
  			u16	attr1; 			// size(2)--  -vflip(1)-  -hflip(1)--  --------rotindex(3)-------  --------x(9)---------
  			u16	attr2; 			// -----------palette(4)-------------  --priority(2)- ---------characterno(10)----------
		  	u16	_rotdata;
		};
		struct
		{
			// attr0 -----------
			u32 y           :8; // Y Coordinate
			u32 rotate      :1;	// Rotate Flag
			u32 doublesize  :1;	// Double Flag
			u32 mode	    :2;	// Mode (Normal, Transparent, etc...)
			u32 mosaic      :1; // Mosaic
			u32 color256    :1; // Color Mode
			u32	shape       :2; // Shape
			// attr1 -----------
			u32 x           :9; // X Coordinate
			/* FIXME: the attempt below doesn't work, see why
			union
			{
			    struct
			    {
			    u32 rotindex:5; // Rotation Index
			    };
			    struct
			    {
			    u32 _unused :3; // Unused bits
			    u32 hflip   :1; // Horizontal Flip
			    u32 vflip   :1; // Vertical Flip
			    };
			};
			*/
			u32 rotindex    :3; // Rotation Index   (this is 5 bits wide!)
			u32 hflip       :1; // Horizontal Flip
			u32 vflip       :1; // Vertical Flip
			u32 size        :2; // Size
			// attr2 -----------
			u32 charno     :10; // Character Number
			u32 priority    :2; // Priority
			u32 palette     :4; // Palette Number (in 16 colors)
			// _rotdata --------
			u16 _rotdata;		// Interleaved rotation data
		};
	};
}			t_sprite_oam;

typedef struct
{
	u16 	_unused1[3];	
	u16 	pa;
	u16 	_unused2[3];	
	u16 	pb;
	u16 	_unused3[3];	
	u16 	pc;
	u16 	_unused4[3];	
	u16 	pd;
}			t_sprite_oam_rotdata;

//---------------------------------------------------------------------
// HIGH-LEVEL SPRITES STRUCTURES
//---------------------------------------------------------------------

typedef struct
{
	t_sprite_oam	oam;					// Actual OAM sprites
	u8				sub_priority;			// Sub Priority (compared to other sprites) (0 = unused sprite)
	u8				next;					// Index of next sprite (in ordered table)
	union
	{
		struct
		{
			s8		x_offset;				// X Coordinate Offset (applied to OAM 'x')
			s8		y_offset;				// Y Coordinate Offset (applied to OAM 'y')
		};
		u16			offsets;
	};
}					t_sprite;

//---------------------------------------------------------------------
// SPRITES DATA
//---------------------------------------------------------------------

// List of sprites (sorted by sub_priority). First cell is a fixed start.
extern t_sprite		sprites[SPR_MAX+1];		

// [Shadow] OAM
extern t_sprite_oam	sprites_oam[SPR_MAX];

// [Shadow] OAM Rotation Data
#define				sprites_oam_rotdata		((t_sprite_oam_rotdata *)sprites_oam)

//---------------------------------------------------------------------
// OAM SPRITES FUNCTIONS
//---------------------------------------------------------------------

// FUNCTION: sprites_oam_clear()
// Clear shadowed OAM and hardware mapped OAM
void		sprites_oam_clear(void);

// FUNCTION: sprites_oam_copy()
// Copy shadowed OAM to hardware mapped OAM
void		sprites_oam_copy(void);

// FUNCTION: sprite_oam_new()
// Return index of first unused sprite in OAM
int			sprite_oam_new(void);

// FUNCTION: sprite_oam_delete()
// Set given sprite to unused state
void		sprite_oam_delete(int idx);

// FUNCTION: sprite_oam_rotate()
// Setup rotation data for given index
void 		sprite_oam_rotate(int rotdata_idx, int angle);

// FUNCTION: sprite_oam_rotate_scale()
// Setup rotation & scaling data for given index
void 		sprite_oam_rotate_scale(int rotdata_idx, int angle, fu16 x_scale, fu16 y_scale);

//---------------------------------------------------------------------
// HIGH-LEVEL SPRITES FUNCTIONS
//---------------------------------------------------------------------

// FUNCTION: sprites_clear()
// Clear OAM and high-level sprite list
void		sprites_clear(void);

// FUNCTION: sprites_copy()
// Copy high-level sprites to hardware mapped OAM, passing thru shadowed OAM
void		sprites_copy(void);

// FUNCTION: sprite_new()
// Return index of first unused sprite
// Must provide sprite sub_priority
int			sprite_new(int sub_priority);

// FUNCTION: sprite_delete()
// Set given sprite to unused state
void		sprite_delete(int idx);

// FUNCTION: sprite_priority_set()
// Change sub_priority of given sprite
void		sprite_priority_set(int idx, int sub_priority);

//---------------------------------------------------------------------
